/*
 * Copyright notice from original mutt:
 * Copyright (C) 1996-2002 Michael R. Elkins <me@mutt.org>
 *
 * This file is part of mutt-ng, see http://www.muttng.org/.
 * It's licensed under the GNU General Public License,
 * please see the file GPL in the top level source directory.
 */

#include <lib-lib/lib-lib.h>

#include <lib-mime/mime.h>
#include <lib-mime/rfc3676.h>
#include <lib-sys/unix.h>
#include <lib-ui/curses.h>
#include <lib-ui/enter.h>
#include <lib-mx/mx.h>

#include "alias.h"
#include "keymap.h"
#include "copy.h"
#include "crypt.h"
#include "mutt_idna.h"
#include "attach.h"

#ifdef USE_NNTP
#include "nntp.h"
#endif

#include "remailer.h"

int url_parse_mailto(ENVELOPE *e, char **body, const char *src)
{
    char *t;
    char *tmp;
    char *headers;
    char *tag, *value;
    char scratch[HUGE_STRING];

    int taglen;

    string_list_t **last = &e->userhdrs;

    if (!(t = strchr (src, ':')))
        return -1;

    if ((tmp = m_strdup(t + 1)) == NULL)
        return -1;

    if ((headers = strchr (tmp, '?')))
        *headers++ = '\0';

    url_decode(tmp);
    e->to = rfc822_parse_adrlist(e->to, tmp);

    tag = headers ? strtok (headers, "&") : NULL;

    for (; tag; tag = strtok(NULL, "&")) {
        if ((value = strchr (tag, '=')))
            *value++ = '\0';
        if (!value || !*value)
            continue;

        url_decode (tag);
        url_decode (value);

        if (mime_which_token(tag, -1) == MIME_BODY) {
            if (body)
                m_strreplace(body, value);
        } else {
#define SAFEPFX (option(OPTSTRICTMAILTO) ? "" : "X-Mailto-")
            taglen = m_strlen(tag) + strlen(SAFEPFX);
            /* mutt_parse_rfc822_line makes some assumptions */
            snprintf(scratch, sizeof(scratch), "%s%s: %s", SAFEPFX, tag, value);
#undef SAVEPFX
            scratch[taglen] = '\0';
            value = vskipspaces(&scratch[taglen + 1]);
            last  = mutt_parse_rfc822_line (e, NULL, scratch, value, 0, 0, last);
            /* if $strict_mailto is set, force editing headers to let
             * users have a look at what we got */
            if (!option (OPTSTRICTMAILTO)) {
                set_option (OPTXMAILTO);
                set_option (OPTEDITHDRS);
            }
        }
    }

    p_delete(&tmp);
    return 0;
}
static void append_signature (FILE * f)
{
  FILE *tmpfp;
  pid_t thepid;

  if (SignOffString) {
    fprintf (f, "\n%s", SignOffString);
  }

  if ((tmpfp = mutt_open_read(NONULL(MAlias.signature), &thepid))) {
    if (option (OPTSIGDASHES))
      fputs ("\n-- \n", f);
    else if (SignOffString)
      fputs ("\n", f);
    mutt_copy_stream (tmpfp, f);
    m_fclose(&tmpfp);
    if (thepid != -1)
      mutt_wait_filter (thepid);
  }
}

/* compare two e-mail addresses and return 1 if they are equivalent */
static int mutt_addrcmp (address_t * a, address_t * b)
{
  if (!a->mailbox || !b->mailbox)
    return 0;
  if (ascii_strcasecmp (a->mailbox, b->mailbox))
    return 0;
  return 1;
}

/* search an e-mail address in a list */
static int mutt_addrsrc (address_t * a, address_t * lst)
{
  for (; lst; lst = lst->next) {
    if (mutt_addrcmp (a, lst))
      return (1);
  }
  return (0);
}

/* removes addresses from "b" which are contained in "a" */
static address_t *mutt_remove_xrefs (address_t * a, address_t * b)
{
  address_t *top, *p, *prev = NULL;

  top = b;
  while (b) {
    for (p = a; p; p = p->next) {
      if (mutt_addrcmp (p, b))
        break;
    }
    if (p) {
      if (prev) {
        prev->next = b->next;
        b->next = NULL;
        address_list_wipe(&b);
        b = prev;
      }
      else {
        top = top->next;
        b->next = NULL;
        address_list_wipe(&b);
        b = top;
      }
    }
    else {
      prev = b;
      b = b->next;
    }
  }
  return top;
}

/* remove any address which matches the current user.  if `leave_only' is
 * nonzero, don't remove the user's address if it is the only one in the list
 */
static address_t *remove_user (address_t * a, int leave_only)
{
  address_t *top = NULL, *last = NULL;

  while (a) {
    if (!mutt_addr_is_user (a)) {
      if (top) {
        last->next = a;
        last = last->next;
      }
      else
        last = top = a;
      a = a->next;
      last->next = NULL;
    }
    else {
      address_t *tmp = a;

      a = a->next;
      if (!leave_only || a || last) {
        tmp->next = NULL;
        address_list_wipe(&tmp);
      }
      else
        last = top = tmp;
    }
  }
  return top;
}

static address_t *find_mailing_lists (address_t * t, address_t * c)
{
  address_t *top = NULL, *ptr = NULL;

  for (; t || c; t = c, c = NULL) {
    for (; t; t = t->next) {
      if (mutt_is_mail_list (t) && !t->group) {
        if (top) {
          ptr->next = address_dup (t);
          ptr = ptr->next;
        }
        else
          ptr = top = address_dup (t);
      }
    }
  }
  return top;
}

static int edit_address (address_t ** a, const char *field)
{
  char buf[HUGE_STRING];
  char *err = NULL;
  int idna_ok = 0;

  do {
    buf[0] = 0;
    mutt_addrlist_to_local (*a);
    rfc822_addrcat(buf, sizeof(buf), *a, 0);
    if (mutt_get_field (field, buf, sizeof (buf), M_ALIAS) != 0)
      return (-1);
    address_list_wipe(a);
    *a = mutt_expand_aliases (mutt_parse_adrlist (NULL, buf));
    if ((idna_ok = mutt_addrlist_to_idna (*a, &err)) != 0) {
      mutt_error (_("Error: '%s' is a bad IDN."), err);
      mutt_refresh ();
      mutt_sleep (2);
      p_delete(&err);
    }
  }
  while (idna_ok != 0);
  return 0;
}

static int edit_envelope (ENVELOPE * en, int flags)
{
  char buf[HUGE_STRING];
  string_list_t *uh = UserHeader;
  regmatch_t pat_match[1];

#ifdef USE_NNTP
  if (option (OPTNEWSSEND)) {
    if (en->newsgroups)
      m_strcpy(buf, sizeof(buf), en->newsgroups);
    else
      buf[0] = 0;
    if (mutt_get_field ("Newsgroups: ", buf, sizeof (buf), 0) != 0)
      return (-1);
    p_delete(&en->newsgroups);
    en->newsgroups = m_strdup(buf);

    if (en->followup_to)
      m_strcpy(buf, sizeof(buf), en->followup_to);
    else
      buf[0] = 0;
    if (option (OPTASKFOLLOWUP)
        && mutt_get_field ("Followup-To: ", buf, sizeof (buf), 0) != 0)
      return (-1);
    p_delete(&en->followup_to);
    en->followup_to = m_strdup(buf);
  } else
#endif
  {
    if (edit_address (&en->to, "To: ") == -1 || en->to == NULL)
      return (-1);
    if (option (OPTASKCC) && edit_address (&en->cc, "Cc: ") == -1)
      return (-1);
    if (option (OPTASKBCC) && edit_address (&en->bcc, "Bcc: ") == -1)
      return (-1);
  }

  if (en->subject) {
    if (option (OPTFASTREPLY))
      return (0);
    else
      m_strcpy(buf, sizeof(buf), en->subject);
  }
  else {
    char *p;

    buf[0] = 0;
    for (; uh; uh = uh->next) {
      if (ascii_strncasecmp ("subject:", uh->data, 8) == 0) {
        p = vskipspaces(uh->data + 8);
        m_strcpy(buf, sizeof(buf), p);
      }
    }
  }

  if ((flags & (SENDREPLY)) && option (OPTSTRIPWAS) && StripWasRegexp.rx &&
      regexec (StripWasRegexp.rx, buf, 1, pat_match, 0) == 0) {
    unsigned int pos = pat_match->rm_so;

    if (ascii_strncasecmp (buf, "re: ", pos) != 0) {
      buf[pos] = '\0';          /* kill match */
      while (pos-- && buf[pos] == ' ')
        buf[pos] = '\0';        /* remove trailing spaces */
    }
    else {
      mutt_error (_("Ignoring $strip_was: Subject would be empty."));
      sleep (2);
    }
  }
  if (mutt_get_field ("Subject: ", buf, sizeof (buf), 0) != 0 || (!buf[0]
                                                                  &&
                                                                  query_quadoption
                                                                  (OPT_SUBJECT,
                                                                   _
                                                                   ("No subject, abort?"))
                                                                  != M_NO)) {
    mutt_message _("No subject, aborting.");

    return (-1);
  }
  m_strreplace(&en->subject, buf);

  return 0;
}

#ifdef USE_NNTP
static char *nntp_get_header(const char *s)
{
    return m_strdup(skipspaces(s));
}
#endif

static void process_user_recips (ENVELOPE * env)
{
    string_list_t *uh = UserHeader;

    for (; uh; uh = uh->next) {
        const char *p = strchr(uh->data, ':');
        if (!p)
            continue;

        switch (mime_which_token(uh->data, p++ - uh->data)) {
          case MIME_TO:
            env->to = rfc822_parse_adrlist(env->to, p);
            break;
          case MIME_CC:
            env->cc = rfc822_parse_adrlist(env->cc, p);
            break;
          case MIME_BCC:
            env->bcc = rfc822_parse_adrlist(env->bcc, p);
            break;
#ifdef USE_NNTP
          case MIME_NEWSGROUPS:
            env->newsgroups = nntp_get_header(p);
            break;
          case MIME_FOLLOWUP_TO:
            env->followup_to = nntp_get_header(p);
            break;
#endif
          default: break;
        }
    }
}

static void process_user_header(ENVELOPE * env)
{
    string_list_t *uh;
    string_list_t **last = string_list_last(&env->userhdrs);

    for (uh = UserHeader; uh; uh = uh->next) {
        const char *p = strchr(uh->data, ':');
        if (!p)
          continue;

        switch (mime_which_token(uh->data, p++ - uh->data)) {
          case MIME_FROM:
            /* User has specified a default From: address.  Remove default address */
            address_list_wipe(&env->from);
            env->from = rfc822_parse_adrlist(env->from, p);
            break;

          case MIME_REPLY_TO:
            address_list_wipe(&env->reply_to);
            env->reply_to = rfc822_parse_adrlist (env->reply_to, p);
            break;

          case MIME_MESSAGE_ID:
            m_strreplace(&env->message_id, p);
            break;

          case MIME_TO:
          case MIME_CC:
          case MIME_BCC:
#ifdef USE_NNTP
          case MIME_NEWSGROUPS:
          case MIME_FOLLOWUP_TO:
#endif
          case MIME_SUPERSEDES:
          case MIME_SUPERCEDES:
          case MIME_SUBJECT:
            break;

          default:
            *last = string_item_new();
            (*last)->data = m_strdup(uh->data);
            last = &(*last)->next;
            break;
        }
    }
}

void mutt_forward_intro (FILE * fp, HEADER * cur)
{
  char buffer[STRING];

  fputs ("----- Forwarded message from ", fp);
  buffer[0] = 0;
  rfc822_addrcat(buffer, sizeof(buffer), cur->env->from, 1);
  fputs (buffer, fp);
  fputs (" -----\n\n", fp);
}

void mutt_forward_trailer (FILE * fp)
{
  fputs ("\n----- End forwarded message -----\n", fp);
}


static int include_forward (CONTEXT * ctx, HEADER * cur, FILE * out)
{
  int chflags = CH_DECODE, cmflags = 0;

  mutt_parse_mime_message (ctx, cur);
  mutt_message_hook (ctx, cur, M_MESSAGEHOOK);

  mutt_forward_intro (out, cur);

  if (option (OPTFORWDECODE)) {
    cmflags |= M_CM_DECODE | M_CM_CHARCONV;
    if (option (OPTWEED)) {
      chflags |= CH_WEED | CH_REORDER;
      cmflags |= M_CM_WEED;
    }
  }
  if (option (OPTFORWQUOTE))
    cmflags |= M_CM_PREFIX;

  mutt_copy_message (out, ctx, cur, cmflags, chflags);
  mutt_forward_trailer (out);
  return 0;
}

void mutt_make_attribution (CONTEXT * ctx, HEADER * cur, FILE * out)
{
  char buffer[STRING];

  if (Attribution) {
    mutt_make_string (buffer, sizeof (buffer), Attribution, ctx, cur);
    fputs (buffer, out);
    fputc ('\n', out);
  }
}

static int include_reply (CONTEXT * ctx, HEADER * cur, FILE * out)
{
  int cmflags = M_CM_PREFIX | M_CM_DECODE | M_CM_CHARCONV | M_CM_REPLYING;
  int chflags = CH_DECODE;

  mutt_parse_mime_message (ctx, cur);
  mutt_message_hook (ctx, cur, M_MESSAGEHOOK);
  mutt_make_attribution (ctx, cur, out);

  if (!option (OPTHEADER))
    cmflags |= M_CM_NOHEADER;
  if (option (OPTWEED)) {
    chflags |= CH_WEED | CH_REORDER;
    cmflags |= M_CM_WEED;
  }

  mutt_copy_message (out, ctx, cur, cmflags, chflags);
  return 0;
}

static int default_to (address_t ** to, ENVELOPE * env, int flags, int hmfupto)
{
  char prompt[STRING];

  if (flags && env->mail_followup_to && hmfupto == M_YES) {
    address_list_append(to, address_list_dup(env->mail_followup_to));
    return 0;
  }

  /* Exit now if we're setting up the default Cc list for list-reply
   * (only set if Mail-Followup-To is present and honoured).
   */
  if (flags & SENDLISTREPLY)
    return 0;

  /* If this message came from a mailing list, ask the user if he really
   * intended to reply to the author only.
   */
  if (!(flags & SENDGROUPREPLY) && mutt_is_list_cc (0, env->to, env->cc)) {
    switch (query_quadoption (OPT_LISTREPLY,
                              _("Message came from a mailing list. List-reply to mailing list?")))
    {
    case M_YES:
      address_list_append(to, find_mailing_lists (env->to, env->cc));
      return 0;
    case -1:
      return -1;                /* abort */
    }
  }

  if (!option (OPTREPLYSELF) && mutt_addr_is_user (env->from)) {
    /* mail is from the user, assume replying to recipients */
    address_list_append(to, address_list_dup(env->to));
  }
  else if (env->reply_to) {
    if ((mutt_addrcmp (env->from, env->reply_to) && !env->reply_to->next) ||
        (option (OPTIGNORELISTREPLYTO) &&
         mutt_is_mail_list (env->reply_to) &&
         (mutt_addrsrc (env->reply_to, env->to) ||
          mutt_addrsrc (env->reply_to, env->cc)))) {
      /* If the Reply-To: address is a mailing list, assume that it was
       * put there by the mailing list, and use the From: address
       * 
       * We also take the from header if our correspondant has a reply-to
       * header which is identical to the electronic mail address given
       * in his From header.
       * 
       */
      address_list_append(to, address_list_dup(env->from));
    }
    else if (!(mutt_addrcmp (env->from, env->reply_to) &&
               !env->reply_to->next) && quadoption (OPT_REPLYTO) != M_YES) {
      /* There are quite a few mailing lists which set the Reply-To:
       * header field to the list address, which makes it quite impossible
       * to send a message to only the sender of the message.  This
       * provides a way to do that.
       */
      snprintf (prompt, sizeof (prompt), _("Reply to %s%s?"),
                env->reply_to->mailbox, env->reply_to->next ? ",..." : "");
      switch (query_quadoption (OPT_REPLYTO, prompt)) {
      case M_YES:
        address_list_append(to, address_list_dup(env->reply_to));
        break;

      case M_NO:
        address_list_append(to, address_list_dup(env->from));
        break;

      default:
        return (-1);            /* abort */
      }
    }
    else
      address_list_append(to, address_list_dup(env->reply_to));
  }
  else
    address_list_append(to, address_list_dup(env->from));

  return (0);
}

int mutt_fetch_recips (ENVELOPE * out, ENVELOPE * in, int flags)
{
  char prompt[STRING];
  int hmfupto = -1;

  if ((flags & (SENDLISTREPLY | SENDGROUPREPLY)) && in->mail_followup_to) {
    snprintf (prompt, sizeof (prompt), _("Follow-up to %s%s?"),
              in->mail_followup_to->mailbox,
              in->mail_followup_to->next ? ",..." : "");

    if ((hmfupto = query_quadoption (OPT_MFUPTO, prompt)) == -1)
      return -1;
  }

  if (flags & SENDLISTREPLY) {
    address_list_append(&out->to, find_mailing_lists(in->to, in->cc));

    if (in->mail_followup_to && hmfupto == M_YES &&
        default_to (&out->cc, in, flags & SENDLISTREPLY, hmfupto) == -1)
      return (-1);              /* abort */
  }
  else {
    if (default_to (&out->to, in, flags & SENDGROUPREPLY, hmfupto) == -1)
      return (-1);              /* abort */

    if ((flags & SENDGROUPREPLY)
        && (!in->mail_followup_to || hmfupto != M_YES))
    {
      address_t **tmp = address_list_append(&out->cc, address_list_dup(in->to));
      address_list_append(tmp, address_list_dup(in->cc));
    }
  }
  return 0;
}

void mutt_fix_reply_recipients (ENVELOPE * env)
{
  mutt_expand_aliases_env (env);

  if (!option (OPTMETOO)) {
    /* the order is important here.  do the CC: first so that if the
     * the user is the only recipient, it ends up on the TO: field
     */
    env->cc = remove_user (env->cc, (env->to == NULL));
    env->to = remove_user (env->to, (env->cc == NULL));
  }

  /* the CC field can get cluttered, especially with lists */
  address_list_uniq(env->to);
  address_list_uniq(env->cc);
  env->cc = mutt_remove_xrefs (env->to, env->cc);

  if (env->cc && !env->to) {
    env->to = env->cc;
    env->cc = NULL;
  }
}

void mutt_make_forward_subject (ENVELOPE * env, CONTEXT * ctx, HEADER * cur)
{
  char buffer[STRING];

  /* set the default subject for the message. */
  mutt_make_string (buffer, sizeof (buffer), NONULL (ForwFmt), ctx, cur);
  m_strreplace(&env->subject, buffer);
}

void mutt_make_misc_reply_headers (ENVELOPE * env,
                                   CONTEXT * ctx __attribute__ ((unused)),
                                   HEADER * cur __attribute__ ((unused)),
                                   ENVELOPE * curenv)
{
  /* This takes precedence over a subject that might have
   * been taken from a List-Post header.  Is that correct?
   */
  if (curenv->real_subj) {
    p_delete(&env->subject);
    env->subject = p_new(char, m_strlen(curenv->real_subj) + 5);
    sprintf (env->subject, "Re: %s", curenv->real_subj);
  }
  else if (!env->subject)
    env->subject = m_strdup("Re: your mail");
}

static string_list_t *mutt_make_references (ENVELOPE * e)
{
  string_list_t *t = NULL, *l = NULL;

  if (e->references)
    l = string_list_dup(e->references);
  else
    l = string_list_dup(e->in_reply_to);

  if (e->message_id) {
    t = string_item_new();
    t->data = m_strdup(e->message_id);
    t->next = l;
    l = t;
  }

  return l;
}

void mutt_add_to_reference_headers (ENVELOPE * env, ENVELOPE * curenv,
                                    string_list_t *** pp, string_list_t *** qq)
{
  string_list_t **p = NULL, **q = NULL;

  if (pp)
    p = *pp;
  if (qq)
    q = *qq;

  if (!p)
    p = &env->references;
  if (!q)
    q = &env->in_reply_to;

  while (*p)
    p = &(*p)->next;
  while (*q)
    q = &(*q)->next;

  *p = mutt_make_references (curenv);

  if (curenv->message_id) {
    *q = string_item_new();
    (*q)->data = m_strdup(curenv->message_id);
  }

  if (pp)
    *pp = p;
  if (qq)
    *qq = q;

}

static void
mutt_make_reference_headers (ENVELOPE * curenv, ENVELOPE * env, CONTEXT * ctx)
{
  env->references = NULL;
  env->in_reply_to = NULL;

  if (!curenv) {
    HEADER *h;
    string_list_t **p = NULL, **q = NULL;
    int i;

    for (i = 0; i < ctx->vcount; i++) {
      h = ctx->hdrs[ctx->v2r[i]];
      if (h->tagged)
        mutt_add_to_reference_headers (env, h->env, &p, &q);
    }
  }
  else
    mutt_add_to_reference_headers (env, curenv, NULL, NULL);
}

static int
envelope_defaults (ENVELOPE * env, CONTEXT * ctx, HEADER * cur, int flags)
{
  ENVELOPE *curenv = NULL;
  int i = 0, tag = 0;

  if (!cur) {
    tag = 1;
    for (i = 0; i < ctx->vcount; i++)
      if (ctx->hdrs[ctx->v2r[i]]->tagged) {
        cur = ctx->hdrs[ctx->v2r[i]];
        curenv = cur->env;
        break;
      }

    if (!cur) {
      /* This could happen if the user tagged some messages and then did
       * a limit such that none of the tagged message are visible.
       */
      mutt_error _("No tagged messages are visible!");

      return (-1);
    }
  }
  else
    curenv = cur->env;

  if (flags & SENDREPLY) {
#ifdef USE_NNTP
    if ((flags & SENDNEWS)) {
      /* in case followup set Newsgroups: with Followup-To: if it present */
      if (!env->newsgroups && curenv &&
          m_strcasecmp(curenv->followup_to, "poster"))
        env->newsgroups = m_strdup(curenv->followup_to);
    }
    else
#endif
    if (tag) {
      HEADER *h;

      for (i = 0; i < ctx->vcount; i++) {
        h = ctx->hdrs[ctx->v2r[i]];
        if (h->tagged && mutt_fetch_recips (env, h->env, flags) == -1)
          return -1;
      }
    }
    else if (mutt_fetch_recips (env, curenv, flags) == -1)
      return -1;

    if ((flags & SENDLISTREPLY) && !env->to) {
      mutt_error _("No mailing lists found!");

      return (-1);
    }

    mutt_make_misc_reply_headers (env, ctx, cur, curenv);
    mutt_make_reference_headers (tag ? NULL : curenv, env, ctx);
  }
  else if (flags & SENDFORWARD)
    mutt_make_forward_subject (env, ctx, cur);

  return (0);
}

static int generate_body (FILE * tempfp,        /* stream for outgoing message */
                          HEADER * msg, /* header for outgoing message */
                          int flags,    /* compose mode */
                          CONTEXT * ctx,        /* current mailbox */
                          HEADER * cur)
{                               /* current message */
  int i;
  HEADER *h;
  BODY *tmp;

  if (flags & SENDREPLY) {
    if ((i =
         query_quadoption (OPT_INCLUDE,
                           _("Include message in reply?"))) == -1)
      return (-1);

    if (i == M_YES) {
      mutt_message _("Including quoted message...");

      if (!cur) {
        for (i = 0; i < ctx->vcount; i++) {
          h = ctx->hdrs[ctx->v2r[i]];
          if (h->tagged) {
            if (include_reply (ctx, h, tempfp) == -1) {
              mutt_error _("Could not include all requested messages!");

              return (-1);
            }
            fputc ('\n', tempfp);
          }
        }
      }
      else
        include_reply (ctx, cur, tempfp);

    }
  }
  else if (flags & SENDFORWARD) {
    if ((i =
         query_quadoption (OPT_MIMEFWD,
                           _("Forward as attachment?"))) == M_YES) {
      BODY *last = msg->content;

      mutt_message _("Preparing forwarded message...");

      while (last && last->next)
        last = last->next;

      if (cur) {
        tmp = mutt_make_message_attach (ctx, cur, 0);
        if (last)
          last->next = tmp;
        else
          msg->content = tmp;
      }
      else {
        for (i = 0; i < ctx->vcount; i++) {
          if (ctx->hdrs[ctx->v2r[i]]->tagged) {
            tmp = mutt_make_message_attach (ctx, ctx->hdrs[ctx->v2r[i]], 0);
            if (last) {
              last->next = tmp;
              last = tmp;
            }
            else
              last = msg->content = tmp;
          }
        }
      }
    }
    else if (i != -1) {
      if (cur)
        include_forward (ctx, cur, tempfp);
      else
        for (i = 0; i < ctx->vcount; i++)
          if (ctx->hdrs[ctx->v2r[i]]->tagged)
            include_forward (ctx, ctx->hdrs[ctx->v2r[i]], tempfp);
    }
    else if (i == -1)
      return -1;
  }

  mutt_clear_error ();

  return (0);
}

void mutt_set_followup_to (ENVELOPE * e)
{
    /*
     * Only generate the Mail-Followup-To if the user has requested it, and
     * it hasn't already been set
     */
    if (!option(OPTFOLLOWUPTO))
        return;

#ifdef USE_NNTP
    if (option(OPTNEWSSEND)) {
        if (!e->followup_to && e->newsgroups && strrchr(e->newsgroups, ','))
            e->followup_to = m_strdup(e->newsgroups);
        return;
    }
#endif

    if (e->mail_followup_to)
        return;

    if (mutt_is_list_cc (0, e->to, e->cc)) {
        address_t **tmp;

        /*
         * this message goes to known mailing lists, so create a proper
         * mail-followup-to header
         */
        tmp = address_list_append(&e->mail_followup_to, address_list_dup(e->to));
        address_list_append(tmp, address_list_dup(e->cc));
    }

    /* remove ourselves from the mail-followup-to header */
    e->mail_followup_to = remove_user(e->mail_followup_to, 0);

    /*
     * If we are not subscribed to any of the lists in question,
     * re-add ourselves to the mail-followup-to header.  The
     * mail-followup-to header generated is a no-op with group-reply,
     * but makes sure list-reply has the desired effect.
     */
    if (e->mail_followup_to && !mutt_is_list_recipient(0, e->to, e->cc)) {
        address_t *from;

        if (e->reply_to)
            from = address_list_dup(e->reply_to);
        else if (e->from)
            from = address_list_dup(e->from);
        else
            from = mutt_default_from();

        address_list_append(&from, e->mail_followup_to);
        e->mail_followup_to = from;
    }

    address_list_uniq(e->mail_followup_to);
}


/* look through the recipients of the message we are replying to, and if
   we find an address that matches $alternates, we use that as the default
   from field */
static address_t *set_reverse_name (ENVELOPE * env)
{
    address_t *tmp = NULL;

    for (tmp = env->to; tmp; tmp = tmp->next) {
        if (mutt_addr_is_user(tmp))
            goto found;
    }
    for (tmp = env->cc; tmp; tmp = tmp->next) {
        if (mutt_addr_is_user(tmp))
            goto found;
    }

    if (!mutt_addr_is_user(env->from))
        return NULL;

    tmp = env->from;

  found:
    tmp = address_dup(tmp);
    if (!option(OPTREVREAL) || !tmp->personal) {
        p_delete(&tmp->personal);
        tmp->personal = m_strdup(Realname);
    }
    return tmp;
}

address_t *mutt_default_from (void)
{
  address_t *adr;

  /*
   * Note: We let $from override $realname here.
   * Is this the right thing to do?
   */

  if (MAlias.from)
    adr = address_dup(MAlias.from);
  else if (mod_core.use_domain) {
    const char *fqdn = mutt_fqdn (1);
    adr = address_new();
    adr->mailbox = p_new(char, m_strlen(mod_core.username) + m_strlen(fqdn) + 2);
    sprintf(adr->mailbox, "%s@%s", NONULL(mod_core.username), NONULL(fqdn));
  } else {
    adr = address_new ();
    adr->mailbox = m_strdup(NONULL(mod_core.username));
  }

  return (adr);
}

static int send_message (HEADER * msg)
{
  char tempfile[_POSIX_PATH_MAX];
  FILE *tempfp;
  int i;

  /* Write out the message in MIME form. */
  tempfp = m_tempfile(tempfile, sizeof(tempfile), NONULL(mod_core.tmpdir), NULL);
  if (!tempfp)
    return -1;

  mutt_write_rfc822_header (tempfp, msg->env, msg->content, 0,
                            msg->chain ? 1 : 0);
  fputc ('\n', tempfp);         /* tie off the header. */

  if ((mutt_write_mime_body (msg->content, tempfp) == -1)) {
    m_fclose(&tempfp);
    unlink (tempfile);
    return (-1);
  }

  if (m_fclose(&tempfp) != 0) {
    mutt_perror (_("Can't create temporary file"));
    unlink (tempfile);
    return (-1);
  }

  if (msg->chain)
    return mix_send_message (msg->chain, tempfile);

  i = mutt_invoke_mta (msg->env->from, msg->env->to, msg->env->cc,
                       msg->env->bcc, tempfile,
                       (msg->content->encoding == ENC8BIT));
  return (i);
}

/* rfc2047 encode the content-descriptions */
static void encode_descriptions (BODY * b, short recurse)
{
  BODY *t;

  for (t = b; t; t = t->next) {
    if (t->description) {
      rfc2047_encode_string (&t->description);
    }
    if (recurse && t->parts)
      encode_descriptions (t->parts, recurse);
  }
}

/* rfc2047 decode them in case of an error */
static void decode_descriptions (BODY * b)
{
  BODY *t;

  for (t = b; t; t = t->next) {
    if (t->description) {
      rfc2047_decode (&t->description);
    }
    if (t->parts)
      decode_descriptions (t->parts);
  }
}

static void fix_end_of_file (const char *data)
{
  FILE *fp;
  int c;

  if ((fp = safe_fopen (data, "a+")) == NULL)
    return;
  fseeko (fp, -1, SEEK_END);
  if ((c = fgetc (fp)) != '\n')
    fputc ('\n', fp);
  m_fclose(&fp);
}

int mutt_resend_message (FILE * fp, CONTEXT * ctx, HEADER * cur)
{
  HEADER *msg = header_new();

  if (mutt_prepare_template (fp, ctx, msg, cur, option(OPTWEED)) < 0)
    return -1;

  return ci_send_message (SENDRESEND, msg, NULL, ctx, cur);
}

int ci_send_message (int flags, /* send mode */
                     HEADER * msg,      /* template to use for new message */
                     char *tempfile,    /* file specified by -i or -H */
                     CONTEXT * ctx,     /* current mailbox */
                     HEADER * cur)
{                               /* current message */
  char fcc[_POSIX_PATH_MAX] = "";       /* where to copy this message */
  FILE *tempfp = NULL;
  BODY *pbody;
  int i, killfrom = 0;
  int fcc_error = 0;
  int free_clear_content = 0;

  BODY *save_content = NULL;
  BODY *clear_content = NULL;
  char *pgpkeylist = NULL;

  /* save current value of "pgp_sign_as" */
  char *signas = NULL, *err = NULL;
  const char *tag = NULL;
  char *ctype;

  int rv = -1;

#ifdef USE_NNTP
  if (flags & SENDNEWS)
    set_option (OPTNEWSSEND);
  else
    unset_option (OPTNEWSSEND);
#endif

  if (!flags && !msg && quadoption (OPT_RECALL) != M_NO &&
      mutt_num_postponed (1)) {
    /* If the user is composing a new message, check to see if there
     * are any postponed messages first.
     */
    if ((i =
         query_quadoption (OPT_RECALL, _("Recall postponed message?"))) == -1)
      return rv;

    if (i == M_YES)
      flags |= SENDPOSTPONED;
  }


  if (flags & SENDPOSTPONED)
    signas = m_strdup(PgpSignAs);

  /* Delay expansion of aliases until absolutely necessary--shouldn't
   * be necessary unless we are prompting the user or about to execute a
   * send-hook.
   */

  if (!msg) {
    msg = header_new();

    if (flags == SENDPOSTPONED) {
      if ((flags =
           mutt_get_postponed (ctx, msg, &cur, fcc, sizeof (fcc))) < 0)
        goto cleanup;
#ifdef USE_NNTP
      /*
       * If postponed message is a news article, it have
       * a "Newsgroups:" header line, then set appropriate flag.
       */
      if (msg->env->newsgroups) {
        flags |= SENDNEWS;
        set_option (OPTNEWSSEND);
      }
      else {
        flags &= ~SENDNEWS;
        unset_option (OPTNEWSSEND);
      }
#endif
    }

    if (flags & (SENDPOSTPONED | SENDRESEND)) {
      if ((tempfp = safe_fopen (msg->content->filename, "a+")) == NULL) {
        mutt_perror (msg->content->filename);
        goto cleanup;
      }
    }

    if (!msg->env)
      msg->env = envelope_new();
  }

  /* Parse and use an eventual list-post header */
  if ((flags & SENDLISTREPLY)
      && cur && cur->env && cur->env->list_post) {
    /* Use any list-post header as a template */
    url_parse_mailto (msg->env, NULL, cur->env->list_post);
    /* We don't let them set the sender's address. */
    address_list_wipe(&msg->env->from);
  }

  if (!(flags & (SENDPOSTPONED | SENDRESEND))) {
    pbody = body_new();
    pbody->next = msg->content; /* don't kill command-line attachments */
    msg->content = pbody;

    if (!(ctype = m_strdup(ContentType)))
      ctype = m_strdup("text/plain");
    mutt_parse_content_type (ctype, msg->content);
    p_delete(&ctype);

    msg->content->unlink = 1;
    msg->content->use_disp = 0;
    msg->content->disposition = DISPINLINE;
    if (option (OPTTEXTFLOWED) && msg->content->type == TYPETEXT
        && !ascii_strcasecmp (msg->content->subtype, "plain")) {
      parameter_setval(&msg->content->parameter, "format", "flowed");
      if (option (OPTDELSP))
        parameter_setval(&msg->content->parameter, "delsp", "yes");
    }

    if (!tempfile) {
      char buffer[_POSIX_PATH_MAX];
      tempfp = m_tempfile(buffer, sizeof(buffer), NONULL(mod_core.tmpdir), NULL);
      msg->content->filename = m_strdup(buffer);
    } else {
      tempfp = safe_fopen(tempfile, "a+");
      msg->content->filename = m_strdup(tempfile);
    }

    if (!tempfp) {
      mutt_perror (msg->content->filename);
      goto cleanup;
    }
  }

  /* this is handled here so that the user can match ~f in send-hook */
  if (cur && option (OPTREVNAME) && !(flags & (SENDPOSTPONED | SENDRESEND))) {
    /* we shouldn't have to worry about freeing `msg->env->from' before
     * setting it here since this code will only execute when doing some
     * sort of reply.  the pointer will only be set when using the -H command
     * line option.
     *
     * We shouldn't have to worry about alias expansion here since we are
     * either replying to a real or postponed message, therefore no aliases
     * should exist since the user has not had the opportunity to add
     * addresses to the list.  We just have to ensure the postponed messages
     * have their aliases expanded.
     */

    msg->env->from = set_reverse_name (cur->env);
  }

  if (!msg->env->from && option (OPTUSEFROM)
      && !(flags & (SENDPOSTPONED | SENDRESEND)))
    msg->env->from = mutt_default_from ();

  if (flags & SENDBATCH) {
    mutt_copy_stream (stdin, tempfp);
    process_user_recips (msg->env);
    process_user_header (msg->env);
    mutt_expand_aliases_env (msg->env);
  }
  else if (!(flags & (SENDPOSTPONED | SENDRESEND))) {
    if ((flags & (SENDREPLY | SENDFORWARD)) && ctx &&
        envelope_defaults (msg->env, ctx, cur, flags) == -1)
      goto cleanup;

    process_user_recips (msg->env);

    /* Expand aliases and remove duplicates/crossrefs */
    mutt_fix_reply_recipients (msg->env);

#ifdef USE_NNTP
    if ((flags & SENDNEWS) && ctx && ctx->magic == M_NNTP
        && !msg->env->newsgroups)
      msg->env->newsgroups = m_strdup(((nntp_data_t *)ctx->data)->group);
#endif

    if (!(option (OPTAUTOEDIT) && option (OPTEDITHDRS)) &&
        !((flags & SENDREPLY) && option (OPTFASTREPLY))) {
      if (edit_envelope (msg->env, flags) == -1)
        goto cleanup;
    }

    /* the from address must be set here regardless of whether or not
     * $use_from is set so that the `~P' (from you) operator in send-hook
     * patterns will work.  if $use_from is unset, the from address is killed
     * after send-hooks are evaulated */

    if (!msg->env->from) {
      msg->env->from = mutt_default_from ();
      killfrom = 1;
    }

    if ((flags & SENDREPLY) && cur) {
      /* change setting based upon message we are replying to */
      mutt_message_hook (ctx, cur, M_REPLYHOOK);

      /*
       * set the replied flag for the message we are generating so that the
       * user can use ~Q in a send-hook to know when reply-hook's are also
       * being used.
       */
      msg->replied = 1;
    }

    /* change settings based upon recipients */

    mutt_message_hook (NULL, msg, M_SENDHOOK);

    /*
     * Unset the replied flag from the message we are composing since it is
     * no longer required.  This is done here because the FCC'd copy of
     * this message was erroneously get the 'R'eplied flag when stored in
     * a maildir-style mailbox.
     */
    msg->replied = 0;

    if (killfrom) {
      address_list_wipe(&msg->env->from);
      killfrom = 0;
    }

    process_user_header (msg->env);

    /* include replies/forwarded messages, unless we are given a template */
    if (!tempfile && (ctx || !(flags & (SENDREPLY | SENDFORWARD)))
        && generate_body (tempfp, msg, flags, ctx, cur) == -1)
      goto cleanup;

    append_signature (tempfp);

    /* 
     * this wants to be done _after_ generate_body, so message-hooks
     * can take effect.
     */

    if (mod_crypt.autosign)
      msg->security |= SIGN;
    if (mod_crypt.autoencrypt)
      msg->security |= ENCRYPT;
    if (mod_crypt.replyencrypt && cur && (cur->security & ENCRYPT))
      msg->security |= ENCRYPT;
    if (mod_crypt.replysign && cur && (cur->security & SIGN))
      msg->security |= SIGN;
    if (mod_crypt.replysignencrypted && cur && (cur->security & ENCRYPT))
      msg->security |= SIGN;

    if (msg->security) {
      /* 
       * When reypling / forwarding, use the original message's
       * crypto system.  According to the documentation,
       * smime_is_default should be disregarded here.
       * 
       * Problem: At least with forwarding, this doesn't really
       * make much sense. Should we have an option to completely
       * disable individual mechanisms at run-time?
       */
      if (cur) {
        if (mod_crypt.autopgp && (cur->security & APPLICATION_PGP))
          msg->security |= APPLICATION_PGP;
        else if (mod_crypt.autosmime && (cur->security & APPLICATION_SMIME))
          msg->security |= APPLICATION_SMIME;
      }

      /*
       * No crypto mechanism selected? Use availability + smime_is_default
       * for the decision. 
       */
      if (!(msg->security & (APPLICATION_SMIME | APPLICATION_PGP))) {
        if (mod_crypt.autosmime && mod_crypt.smime_is_default)
          msg->security |= APPLICATION_SMIME;
        else if (mod_crypt.autopgp)
          msg->security |= APPLICATION_PGP;
        else if (mod_crypt.autosmime)
          msg->security |= APPLICATION_SMIME;
      }
    }

    /* No permissible mechanisms found.  Don't sign or encrypt. */
    if (!(msg->security & (APPLICATION_SMIME | APPLICATION_PGP)))
      msg->security = 0;
  }

  /* 
   * This hook is even called for postponed messages, and can, e.g., be
   * used for setting the editor, the sendmail path, or the
   * envelope sender.
   */
  mutt_message_hook (NULL, msg, M_SEND2HOOK);

  /* wait until now to set the real name portion of our return address so
     that $realname can be set in a send-hook */
  if (msg->env->from && !msg->env->from->personal
      && !(flags & (SENDRESEND | SENDPOSTPONED)))
    msg->env->from->personal = m_strdup(Realname);

  m_fclose(&tempfp);

  if (!(flags & SENDBATCH)) {
    struct stat st;
    time_t mtime = m_decrease_mtime(msg->content->filename, NULL);

    mutt_update_encoding (msg->content);

    /*
     * Select whether or not the user's editor should be called now.  We
     * don't want to do this when:
     * 1) we are sending a key/cert
     * 2) we are forwarding a message and the user doesn't want to edit it.
     *    This is controled by the quadoption $forward_edit.  However, if
     *    both $edit_headers and $autoedit are set, we want to ignore the
     *    setting of $forward_edit because the user probably needs to add the
     *    recipients.
     */
    if (((flags & SENDFORWARD) == 0 ||
         (option (OPTEDITHDRS) && option (OPTAUTOEDIT)) ||
         query_quadoption (OPT_FORWEDIT,
                           _("Edit forwarded message?")) == M_YES))
    {
      /* If the this isn't a text message, look for a mailcap edit command */
      if (rfc1524_mailcap_isneeded(msg->content)) {
        if (!mutt_edit_attachment (msg->content))
          goto cleanup;
      } else if (option (OPTEDITHDRS)) {
        mutt_env_to_local (msg->env);
        mutt_edit_headers(msg->content->filename, msg, fcc, sizeof (fcc));
        mutt_env_to_idna (msg->env, NULL, NULL);
      }
      else {
        mutt_edit_file(msg->content->filename);

        if (stat (msg->content->filename, &st) == 0) {
          if (mtime != st.st_mtime)
            fix_end_of_file (msg->content->filename);
        } else
          mutt_perror (msg->content->filename);
      }

      if (option (OPTTEXTFLOWED))
        rfc3676_space_stuff (msg);

      mutt_message_hook (NULL, msg, M_SEND2HOOK);
    }

    if (!(flags & (SENDPOSTPONED | SENDFORWARD | SENDRESEND))) {
      if (stat (msg->content->filename, &st) == 0) {
        /* if the file was not modified, bail out now */
        if (mtime == st.st_mtime && !msg->content->next &&
            query_quadoption (OPT_ABORT,
                              _("Abort unmodified message?")) == M_YES) {
          mutt_message _("Aborted unmodified message.");

          goto cleanup;
        }
      }
      else
        mutt_perror (msg->content->filename);
    }
  }

  /* specify a default fcc.  if we are in batchmode, only save a copy of
   * the message if the value of $copy is yes or ask-yes */

  if (!fcc[0] && !(flags & (SENDPOSTPONED))
      && (!(flags & SENDBATCH) || (quadoption (OPT_COPY) & 0x1))) {
    /* set the default FCC */
    if (!msg->env->from) {
      msg->env->from = mutt_default_from ();
      killfrom = 1;             /* no need to check $use_from because if the user specified
                                   a from address it would have already been set by now */
    }
    mutt_select_fcc (fcc, sizeof (fcc), msg);
    if (killfrom) {
      address_list_wipe(&msg->env->from);
      killfrom = 0;
    }
  }


  mutt_update_encoding (msg->content);

  if (!(flags & SENDBATCH)) {
  main_loop:

    fcc_error = 0;              /* reset value since we may have failed before */
    mutt_pretty_mailbox (fcc);
    i = mutt_compose_menu (msg, fcc, sizeof (fcc), cur);
    if (i == -1) {
      /* abort */
#ifdef USE_NNTP
      if (flags & SENDNEWS)
        mutt_message (_("Article not posted."));

      else
#endif
        mutt_message _("Mail not sent.");
      goto cleanup;
    }
    else if (i == 1) {
      /* postpone the message until later. */
      if (msg->content->next)
        msg->content = mutt_make_multipart (msg->content);

      /*
       * make sure the message is written to the right part of a maildir 
       * postponed folder.
       */
      msg->read = 0;
      msg->old = 0;

      encode_descriptions (msg->content, 1);
      mutt_prepare_envelope (msg->env, 0);
      mutt_env_to_idna (msg->env, NULL, NULL);  /* Handle bad IDNAs the next time. */

      if (!Postponed
          || mutt_write_fcc (NONULL (Postponed), msg,
                             (cur
                              && (flags & SENDREPLY)) ? cur->env->
                             message_id : NULL, 1, fcc) < 0) {
        msg->content = mutt_remove_multipart (msg->content);
        decode_descriptions (msg->content);
        mutt_unprepare_envelope (msg->env);
        goto main_loop;
      }
      mutt_update_num_postponed ();
      mutt_message _("Message postponed.");

      goto cleanup;
    }
  }

#ifdef USE_NNTP
  if (!(flags & SENDNEWS))
#endif
    if (!msg->env->to && !msg->env->cc && !msg->env->bcc) {
      if (!(flags & SENDBATCH)) {
        mutt_error _("No recipients are specified!");

        goto main_loop;
      }
      else {
        puts _("No recipients were specified.");

        goto cleanup;
      }
    }

  if (mutt_env_to_idna (msg->env, &tag, &err)) {
    mutt_error (_("Bad IDN in \"%s\": '%s'"), tag, err);
    p_delete(&err);
    if (!(flags & SENDBATCH))
      goto main_loop;
    else
      goto cleanup;
  }

  if (!msg->env->subject && !(flags & SENDBATCH) &&
      (i =
       query_quadoption (OPT_SUBJECT,
                         _("No subject, abort sending?"))) != M_NO) {
    /* if the abort is automatic, print an error message */
    if (quadoption (OPT_SUBJECT) == M_YES)
      mutt_error _("No subject specified.");

    goto main_loop;
  }
#ifdef USE_NNTP
  if ((flags & SENDNEWS) && !msg->env->subject) {
    mutt_error _("No subject specified.");

    goto main_loop;
  }

  if ((flags & SENDNEWS) && !msg->env->newsgroups) {
    mutt_error _("No newsgroup specified.");

    goto main_loop;
  }
#endif

  if (msg->content->next)
    msg->content = mutt_make_multipart (msg->content);

  if (mutt_attach_check (msg) &&
      !msg->content->next &&
      query_quadoption (OPT_ATTACH,
                        _("No attachments made but indicator found in text. "
                          "Cancel sending?")) == M_YES) {
    if (quadoption (OPT_ATTACH) == M_YES) {
      mutt_message _("No attachments made but indicator found in text. "
                     "Abort sending.");
      sleep (2);
    }
    mutt_message (_("Mail not sent."));
    goto main_loop;
  }

  /* 
   * Ok, we need to do it this way instead of handling all fcc stuff in
   * one place in order to avoid going to main_loop with encoded "env"
   * in case of error.  Ugh.
   */

  encode_descriptions (msg->content, 1);

  /*
   * Make sure that clear_content and free_clear_content are
   * properly initialized -- we may visit this particular place in
   * the code multiple times, including after a failed call to
   * mutt_protect().
   */

  clear_content = NULL;
  free_clear_content = 0;

  if (msg->security) {
    /* save the decrypted attachments */
    clear_content = msg->content;

    if ((crypt_get_keys (msg, &pgpkeylist) == -1) ||
        mutt_protect (msg, pgpkeylist) == -1) {
      msg->content = mutt_remove_multipart (msg->content);

      p_delete(&pgpkeylist);

      decode_descriptions (msg->content);
      goto main_loop;
    }
    encode_descriptions (msg->content, 0);
  }

  /* 
   * at this point, msg->content is one of the following three things:
   * - multipart/signed.  In this case, clear_content is a child.
   * - multipart/encrypted.  In this case, clear_content exists
   *   independently
   * - application/pgp.  In this case, clear_content exists independently.
   * - something else.  In this case, it's the same as clear_content.
   */

  /* This is ugly -- lack of "reporting back" from mutt_protect(). */

  if (clear_content && (msg->content != clear_content)
      && (msg->content->parts != clear_content))
    free_clear_content = 1;

  if (!option (OPTNOCURSES))
    mutt_message _("Sending message...");

  mutt_prepare_envelope (msg->env, 1);

  /* save a copy of the message, if necessary. */

  mutt_expand_path (fcc, sizeof (fcc));


  /* Don't save a copy when we are in batch-mode, and the FCC
   * folder is on an IMAP server: This would involve possibly lots
   * of user interaction, which is not available in batch mode. 
   * 
   * Note: A patch to fix the problems with the use of IMAP servers
   * from non-curses mode is available from Brendan Cully.  However, 
   * I'd like to think a bit more about this before including it.
   */

  if ((flags & SENDBATCH) && fcc[0] && mx_get_magic (fcc) == M_IMAP)
    fcc[0] = '\0';

  if (*fcc && m_strcmp("/dev/null", fcc) != 0) {
    BODY *tmpbody = msg->content;
    BODY *save_sig = NULL;
    BODY *save_parts = NULL;

    if (msg->security && option (OPTFCCCLEAR))
      msg->content = clear_content;

    /* check to see if the user wants copies of all attachments */
    if (!option (OPTFCCATTACH) && msg->content->type == TYPEMULTIPART) {
      if ((m_strcmp(msg->content->subtype, "encrypted") == 0 ||
              m_strcmp(msg->content->subtype, "signed") == 0))
      {
        if (clear_content->type == TYPEMULTIPART) {
          if (!(msg->security & ENCRYPT) && (msg->security & SIGN)) {
            /* save initial signature and attachments */
            save_sig = msg->content->parts->next;
            save_parts = clear_content->parts->next;
          }

          /* this means writing only the main part */
          msg->content = clear_content->parts;

          if (mutt_protect (msg, pgpkeylist) == -1) {
            /* we can't do much about it at this point, so
             * fallback to saving the whole thing to fcc
             */
            msg->content = tmpbody;
            save_sig = NULL;
            goto full_fcc;
          }

          save_content = msg->content;
        }
      }
      else
        msg->content = msg->content->parts;
    }

  full_fcc:
    if (msg->content) {
      /* update received time so that when storing to a mbox-style folder
       * the From_ line contains the current time instead of when the
       * message was first postponed.
       */
      msg->received = time (NULL);
      if (mutt_write_fcc (fcc, msg, NULL, 0, NULL) == -1) {
        /*
         * Error writing FCC, we should abort sending.
         */
        fcc_error = 1;
      }
    }

    msg->content = tmpbody;

    if (save_sig) {
      /* cleanup the second signature structures */
      if (save_content->parts) {
        body_list_wipe(&save_content->parts->next);
        save_content->parts = NULL;
      }
      body_list_wipe(&save_content);

      /* restore old signature and attachments */
      msg->content->parts->next = save_sig;
      msg->content->parts->parts->next = save_parts;
    }
    else if (save_content) {
      /* destroy the new encrypted body. */
      body_list_wipe(&save_content);
    }

  }


  /*
   * Don't attempt to send the message if the FCC failed.  Just pretend
   * the send failed as well so we give the user a chance to fix the
   * error.
   */
  if (fcc_error || (i = send_message (msg)) == -1) {
    if (!(flags & SENDBATCH)) {
      if ((msg->security & ENCRYPT)
      ||  ((msg->security & SIGN)
      &&  msg->content->type == TYPEAPPLICATION)) {
        body_list_wipe(&msg->content); /* destroy PGP data */
        msg->content = clear_content;   /* restore clear text. */
      }
      else if ((msg->security & SIGN) && msg->content->type == TYPEMULTIPART) {
        body_list_wipe(&msg->content->parts->next);    /* destroy sig */
        msg->content = mutt_remove_multipart (msg->content);
      }

      msg->content = mutt_remove_multipart (msg->content);
      decode_descriptions (msg->content);
      mutt_unprepare_envelope (msg->env);
      goto main_loop;
    }
    else {
      puts _("Could not send the message.");

      goto cleanup;
    }
  }
  else if (!option (OPTNOCURSES))
    mutt_message (i != 0 ? _("Sending in background.") :
#ifdef USE_NNTP
                  (flags & SENDNEWS) ? _("Article posted.") :
                  _("Mail sent.")
#else
                  _("Mail sent.")
#endif
    );
  if (msg->security & ENCRYPT)
    p_delete(&pgpkeylist);

  if (free_clear_content)
    body_list_wipe(&clear_content);

  if (flags & SENDREPLY) {
    if (cur && ctx)
      mutt_set_flag (ctx, cur, M_REPLIED, 1);
    else if (!(flags & SENDPOSTPONED) && ctx && ctx->tagged) {
      for (i = 0; i < ctx->vcount; i++)
        if (ctx->hdrs[ctx->v2r[i]]->tagged)
          mutt_set_flag (ctx, ctx->hdrs[ctx->v2r[i]], M_REPLIED, 1);
    }
  }


  rv = 0;

cleanup:

  if (flags & SENDPOSTPONED) {
    if (signas) {
      p_delete(&PgpSignAs);
      PgpSignAs = signas;
    }
  }

  m_fclose(&tempfp);
  header_delete(&msg);

  return rv;
}

/* vim: set sw=2: */
